@@ -34,6 +34,7 @@
 #include <gl_linked_list.h>
 #include <gl_xlist.h>
 #include <minmax.h>
+#include <safe.h>
 
 /* procedures */
 
@@ -114,6 +115,7 @@ main (int argc, char **argv)
     struct stat tmpoutst;
     char numbuf[LINENUM_LENGTH_BOUND + 1];
     bool written_to_rejname = false;
+    bool skip_reject_file = false;
     bool apply_empty_patch = false;
     mode_t file_type;
     int outfd = -1;
@@ -189,6 +191,7 @@ main (int argc, char **argv)
 	there_is_another_patch (! (inname || posixly_correct), &file_type)
 	  || apply_empty_patch;
 	reinitialize_almost_everything(),
+	  skip_reject_file = false,
 	  apply_empty_patch = false
     ) {					/* for each patch in patch file */
       int hunk = 0;
@@ -243,7 +246,7 @@ main (int argc, char **argv)
 	  if (outfile)
 	    outname = outfile;
 	  else if (pch_copy () || pch_rename ())
-	    outname = pch_name (! strcmp (inname, pch_name (OLD)));
+	    outname = pch_name (! reverse);
 	  else
 	    outname = inname;
 	}
@@ -291,7 +294,7 @@ main (int argc, char **argv)
 
       if (read_only_behavior != RO_IGNORE
 	  && ! inerrno && ! S_ISLNK (instat.st_mode)
-	  && access (inname, W_OK) != 0)
+	  && safe_access (inname, W_OK) != 0)
 	{
 	  say ("File %s is read-only; ", quotearg (inname));
 	  if (read_only_behavior == RO_WARN)
@@ -308,7 +311,20 @@ main (int argc, char **argv)
       outfd = make_tempfile (&TMPOUTNAME, 'o', outname,
 			     O_WRONLY | binary_transput,
 			     instat.st_mode & S_IRWXUGO);
-      TMPOUTNAME_needs_removal = true;
+      if (outfd == -1)
+	{
+	  if (errno == ELOOP || errno == EXDEV)
+	    {
+	      say ("Invalid file name %s -- skipping patch\n", quotearg (outname));
+	      skip_rest_of_patch = true;
+	      skip_reject_file = true;
+	      somefailed = true;
+	    }
+	  else
+	    pfatal ("Can't create temporary file %s", TMPOUTNAME);
+	}
+      else
+        TMPOUTNAME_needs_removal = true;
       if (diff_type == ED_DIFF) {
 	outstate.zero_output = false;
 	somefailed |= skip_rest_of_patch;
@@ -339,6 +355,8 @@ main (int argc, char **argv)
 	    outstate.ofp = fdopen(outfd, binary_transput ? "wb" : "w");
 	    if (! outstate.ofp)
 	      pfatal ("%s", TMPOUTNAME);
+	    /* outstate.ofp now owns the file descriptor */
+	    outfd = -1;
 	  }
 
 	/* find out where all the lines are */
@@ -348,16 +366,18 @@ main (int argc, char **argv)
 	    if (verbosity != SILENT)
 	      {
 		bool renamed = strcmp (inname, outname);
+		bool skip_rename = ! renamed && pch_rename ();
 
 		say ("%s %s %s%c",
 		     dry_run ? "checking" : "patching",
 		     S_ISLNK (file_type) ? "symbolic link" : "file",
-		     quotearg (outname), renamed ? ' ' : '\n');
-		if (renamed)
-		  say ("(%s from %s)\n",
+		     quotearg (outname), renamed || skip_rename ? ' ' : '\n');
+		if (renamed || skip_rename)
+		  say ("(%s%s from %s)\n",
+		       skip_rename ? "already " : "",
 		       pch_copy () ? "copied" :
 		       (pch_rename () ? "renamed" : "read"),
-		       inname);
+		       ! skip_rename ? inname : pch_name (! strcmp (inname, pch_name (OLD))));
 		if (verbosity == VERBOSE)
 		  say ("Using Plan %s...\n", using_plan_a ? "A" : "B");
 	      }
@@ -447,7 +467,8 @@ main (int argc, char **argv)
 			|| ! where
 			|| ! apply_hunk (&outstate, where))))
 	      {
-		abort_hunk (outname, ! failed, reverse);
+		if (! skip_reject_file)
+		  abort_hunk (outname, ! failed, reverse);
 		failed++;
 		if (verbosity == VERBOSE ||
 		    (! skip_rest_of_patch && verbosity != SILENT))
@@ -540,7 +561,7 @@ main (int argc, char **argv)
 		      enum file_attributes attr = 0;
 		      struct timespec new_time = pch_timestamp (! reverse);
 		      mode_t mode = file_type |
-			  ((new_mode ? new_mode : instat.st_mode) & S_IRWXUGO);
+			  ((set_mode ? new_mode : instat.st_mode) & S_IRWXUGO);
 
 		      if ((set_time | set_utc) && new_time.tv_sec != -1)
 			{
@@ -588,7 +609,7 @@ main (int argc, char **argv)
       if (diff_type != ED_DIFF) {
 	struct stat rejst;
 
-	if (failed) {
+	if (failed && ! skip_reject_file) {
 	    if (fstat (fileno (rejfp), &rejst) != 0 || fclose (rejfp) != 0)
 	      write_fatal ();
 	    rejfp = NULL;
@@ -1581,6 +1602,8 @@ init_reject (char const *outname)
   int fd;
   fd = make_tempfile (&TMPREJNAME, 'r', outname, O_WRONLY | binary_transput,
 		      0666);
+  if (fd == -1)
+    pfatal ("Can't create temporary file %s", TMPREJNAME);
   TMPREJNAME_needs_removal = true;
   rejfp = fdopen (fd, binary_transput ? "wb" : "w");
   if (! rejfp)
@@ -1917,7 +1940,7 @@ output_files (struct stat const *st)
 		       from_st, file_to_output->to,
 		       file_to_output->mode, file_to_output->backup);
       if (file_to_output->to && from_needs_removal)
-	unlink (file_to_output->from);
+	safe_unlink (file_to_output->from);
 
       if (st && st->st_dev == from_st->st_dev && st->st_ino == from_st->st_ino)
 	{
@@ -1947,7 +1970,7 @@ forget_output_files (void)
     {
       const struct file_to_output *file_to_output = elt;
 
-      unlink (file_to_output->from);
+      safe_unlink (file_to_output->from);
     }
   gl_list_iterator_free (&iter);
   gl_list_clear (files_to_output);
@@ -1971,7 +1994,7 @@ remove_if_needed (char const *name, bool *needs_removal)
 {
   if (*needs_removal)
     {
-      unlink (name);
+      safe_unlink (name);
       *needs_removal = false;
     }
 }
